/* eslint-disable no-magic-numbers */
import { Style, Text } from 'ol/style';
import Feature, { FeatureLike } from 'ol/Feature';
import { MultiPolygon } from 'ol/geom';
import { UsePointToProjectionResult } from '@/utils/map/usePointToProjection';
import getStroke from '@/components/Map/MapViewer/MapViewerVector/utils/shapes/getStroke';
import getFill from '@/components/Map/MapViewer/MapViewerVector/utils/shapes/getFill';
import Polygon from 'ol/geom/Polygon';
import getMultiPolygon from '@/components/Map/MapViewer/MapViewerVector/utils/shapes/getMultiPolygon';
import getText from '@/components/Map/MapViewer/MapViewerVector/utils/shapes/getText';
import { rgbToHex } from '@/components/Map/MapViewer/MapViewerVector/utils/shapes/rgbToHex';
import { BioShape, Modification, Shape } from '@/types/models';
import { MapInstance } from '@/types/map';
import {
  ColorObject,
  HorizontalAlign,
  VerticalAlign,
} from '@/components/Map/MapViewer/MapViewerVector/MapViewerVector.types';
import {
  BLACK_COLOR,
  WHITE_COLOR,
} from '@/components/Map/MapViewer/MapViewerVector/MapViewerVector.constants';

export type MapElementProps = {
  shapes: Array<Shape>;
  x: number;
  y: number;
  width: number;
  height: number;
  zIndex: number;
  fillColor?: ColorObject;
  borderColor?: ColorObject;
  fontColor?: ColorObject;
  lineWidth?: number;
  text?: string;
  fontSize?: string | number;
  nameX: number;
  nameY: number;
  nameHeight: number;
  nameWidth: number;
  nameVerticalAlign?: VerticalAlign;
  nameHorizontalAlign?: HorizontalAlign;
  pointToProjection: UsePointToProjectionResult;
  mapInstance: MapInstance;
  bioShapes?: Array<BioShape>;
  modifications?: Array<Modification>;
};

export default class MapElement {
  textStyle: Style | undefined;

  polygons: Array<Polygon> = [];

  styles: Array<Style> = [];

  polygonsTexts: Array<string> = [];

  multiPolygonFeature: Feature;

  constructor({
    shapes,
    x,
    y,
    width,
    height,
    zIndex,
    fillColor = WHITE_COLOR,
    borderColor = BLACK_COLOR,
    fontColor = BLACK_COLOR,
    lineWidth = 1,
    text = '',
    fontSize = 12,
    nameX,
    nameY,
    nameHeight,
    nameWidth,
    nameVerticalAlign = 'MIDDLE',
    nameHorizontalAlign = 'CENTER',
    pointToProjection,
    mapInstance,
    bioShapes = [],
    modifications = [],
  }: MapElementProps) {
    if (text) {
      const { textCoords, textStyle } = getText({
        text,
        fontSize,
        x: nameX,
        y: nameY,
        width: nameWidth,
        height: nameHeight,
        color: rgbToHex(fontColor),
        zIndex,
        verticalAlign: nameVerticalAlign,
        horizontalAlign: nameHorizontalAlign,
        pointToProjection,
      });
      this.styles.push(textStyle);
      this.polygonsTexts.push(text);
      this.polygons.push(new Polygon([[textCoords, textCoords]]));
    }

    const multiPolygon: Array<Polygon> = [];
    modifications.forEach(modification => {
      if (modification.state === null) {
        return;
      }

      const shape = bioShapes.find(bioShape => bioShape.sboTerm === modification.sboTerm);
      if (!shape) {
        return;
      }
      const multiPolygonModification = getMultiPolygon({
        x: modification.x,
        y: modification.y,
        width: modification.width,
        height: modification.height,
        shapes: shape.shapes,
        pointToProjection,
        mirror: modification.direction && modification.direction === 'RIGHT',
      });
      multiPolygon.push(...multiPolygonModification.flat());
      multiPolygonModification.forEach(polygon => {
        const modificationStyle = new Style({
          geometry: polygon,
          stroke: getStroke({ color: rgbToHex(modification.borderColor) }),
          fill: getFill({ color: rgbToHex(modification.fillColor) }),
          zIndex: modification.z,
        });
        const modificationText = modification.stateAbbreviation
          ? modification.stateAbbreviation
          : modification.name;
        if (modificationText) {
          modificationStyle.setText(
            new Text({
              text: modificationText,
              font: `${modification.fontSize}px Arial`,
              textAlign: 'center',
              textBaseline: 'middle',
              fill: getFill({ color: '#000' }),
              overflow: true,
            }),
          );
          this.polygonsTexts.push(modification.name);
        }
        this.styles.push(modificationStyle);
      });
    });

    const elementMultiPolygon = getMultiPolygon({ x, y, width, height, shapes, pointToProjection });
    this.polygons = [...this.polygons, ...multiPolygon, ...elementMultiPolygon.flat()];

    elementMultiPolygon.forEach(polygon => {
      this.styles.push(
        new Style({
          geometry: polygon,
          stroke: getStroke({ color: rgbToHex(borderColor), width: lineWidth }),
          fill: getFill({ color: rgbToHex(fillColor) }),
          zIndex,
        }),
      );
    });

    this.multiPolygonFeature = new Feature({
      geometry: new MultiPolygon([...this.polygons]),
      getTextScale: (resolution: number): number => {
        const maxZoom = mapInstance?.getView().getMaxZoom();
        if (maxZoom) {
          const minResolution = mapInstance?.getView().getResolutionForZoom(maxZoom);
          if (minResolution) {
            return Math.round((minResolution / resolution) * 100) / 100;
          }
        }
        return 1;
      },
    });

    this.multiPolygonFeature.setStyle(this.styleFunction.bind(this));
  }

  styleFunction(feature: FeatureLike, resolution: number): Style | Array<Style> | void {
    const getTextScale = feature.get('getTextScale');
    let textScale = 1;
    if (getTextScale instanceof Function) {
      textScale = getTextScale(resolution);
    }
    let index = 0;
    this.styles.forEach(style => {
      if (style.getText()) {
        if (textScale > 0.3) {
          style.getText()?.setScale(textScale);
          style.getText()?.setText(this.polygonsTexts[index]);
          index += 1;
        } else {
          style.getText()?.setText(undefined);
        }
      }
    });
    return this.styles;
  }
}
